__all__ = [
    "QuartProvider",
]

from dishka import Provider, Scope, from_context
from quart import Request, Websocket


class QuartProvider(Provider):
    request = from_context(Request, scope=Scope.REQUEST)
    websocket = from_context(Websocket, scope=Scope.SESSION)



from typing import Any

from dishka import AsyncContainer, Scope
from quart import Quart, Request, Websocket, g, request, websocket


class ContainerMiddleware:
    def __init__(self, container: AsyncContainer) -> None:
        self.container = container

    def setup(self, app: Quart) -> None:
        app.before_request(self.enter_request)
        app.before_websocket(self.enter_websocket)
        app.teardown_request(self.exit_scope)
        app.teardown_websocket(self.exit_scope)

    async def enter_request(self) -> None:
        wrapper = self.container({Request: request}, scope=Scope.REQUEST)
        g.dishka_container_wrapper = wrapper
        g.dishka_container = await wrapper.__aenter__()

    async def enter_websocket(self) -> None:
        wrapper = self.container({Websocket: websocket}, scope=Scope.SESSION)
        g.dishka_container_wrapper = wrapper
        g.dishka_container = await wrapper.__aenter__()

    # noinspection PyMethodMayBeStatic
    async def exit_scope(self, *_args: Any, **_kwargs: Any) -> None:
        if hasattr(g, "dishka_container"):
            await g.dishka_container.close()



__all__ = [
    "QuartDishka",
    "inject",
]

from collections.abc import Callable
from functools import wraps
from typing import Any, TypeAlias, TypeVar, cast

from dishka import AsyncContainer
from dishka.integrations.base import is_dishka_injected, wrap_injection
from quart import Quart, g
from quart.blueprints import Blueprint
from quart.typing import RouteCallable, WebsocketCallable

from quart_dishka.container import ContainerMiddleware
from quart_dishka.exceptions import ContainerNotSetError

try:
    from typing import ParamSpec
except ImportError:
    from typing_extensions import ParamSpec  # type: ignore[assignment]

P = ParamSpec("P")
T = TypeVar("T")
Scaffold: TypeAlias = Blueprint | Quart


def _inject_routes(app: Scaffold) -> None:
    for endpoint, func in app.view_functions.items():
        if not is_dishka_injected(func):
            wrapped = _make_wrapper(func)
            if getattr(func, "websocket", False):
                # Note: expression includes quart.wrappers.response.Response,
                #       but target expects werkzeug.sansio.response.Response
                app.view_functions[endpoint] = cast(WebsocketCallable, wrapped)  # type: ignore[assignment]
            else:
                app.view_functions[endpoint] = cast(RouteCallable, wrapped)


def _make_wrapper(func: Any) -> Any:
    @wraps(func)
    async def wrapped(*args: Any, **kwargs: Any) -> Any:
        injected = inject(func)
        result = injected(*args, **kwargs)
        if hasattr(result, "__await__"):
            result = await result
        return result

    return wrapped


def inject(func: Callable[P, T]) -> Callable[P, T]:
    return wrap_injection(
        func=func,
        is_async=True,
        container_getter=lambda _, p: g.dishka_container,
    )


class QuartDishka:
    """Quart extension for Dishka dependency injection.

    Example:
        >>> app = Quart(__name__)
        >>> container = make_async_container(QuartProvider())
        >>> QuartDishka(app=app, container=container)

    Args:
        app: Optional Quart application
        container: Optional AsyncContainer instance
        auto_inject: If True, enables auto-injection for all routes
    """

    def __init__(
            self,
            app: Quart | None = None,
            container: AsyncContainer | None = None,
            *,
            auto_inject: bool = False,
    ) -> None:
        self.container = container
        self.auto_inject = auto_inject

        if app is not None:
            self.init_app(app)

    def init_app(self, app: Quart) -> None:

        if not self.container:
            raise ContainerNotSetError

        middleware = ContainerMiddleware(self.container)
        middleware.setup(app)

        if self.auto_inject:
            _inject_routes(app)
            for blueprint in app.blueprints.values():
                # Note: incoming type is flask.sansio.blueprints.Blueprint,
                #       but at runtime it's quart.blueprints.Blueprint
                _inject_routes(blueprint)  # type: ignore[arg-type]

        app.extensions["QUART_DISHKA"] = self



from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager

import pytest
from dishka import FromDishka, make_async_container
from quart import Quart, Request, Websocket

from quart_dishka.container import ContainerMiddleware
from quart_dishka.extension import inject
from quart_dishka.provider import QuartProvider


@asynccontextmanager
async def provider_http_app(handler) -> AsyncGenerator[Quart, None]:
    app = Quart(__name__)
    app.route("/", methods=["GET", "POST"])(inject(handler))
    container = make_async_container(QuartProvider())

    middleware = ContainerMiddleware(container)
    middleware.setup(app)

    try:
        yield app
    finally:
        await container.close()


@asynccontextmanager
async def provider_ws_app(handler) -> AsyncGenerator[Quart, None]:
    app = Quart(__name__)
    app.websocket("/")(inject(handler))
    container = make_async_container(QuartProvider())

    middleware = ContainerMiddleware(container)
    middleware.setup(app)

    try:
        yield app
    finally:
        await container.close()


async def handle_request_get(request: FromDishka[Request]) -> dict:
    return {
        "method": request.method,
        "path": request.path,
        "headers": dict(request.headers),
        "query_string": request.query_string.decode(),
    }


async def handle_request_post(request: FromDishka[Request]) -> dict:
    data = await request.get_json()
    return {
        "method": request.method,
        "content_type": request.headers["content-type"],
        "data": data,
    }


async def handle_websocket(websocket: FromDishka[Websocket]) -> None:
    await websocket.accept()

    headers = dict(websocket.headers)
    path = websocket.path

    data = await websocket.receive_json()

    await websocket.send_json({
        "headers": headers,
        "path": path,
        "echo": data,
    })


@pytest.mark.asyncio
async def test_quart_provider_request_get():
    async with provider_http_app(handle_request_get) as app:
        client = app.test_client()

        headers = {"X-Test": "test-value"}
        response = await client.get("/?param=value", headers=headers)

        assert response.status_code == 200
        data = await response.get_json()

        assert data["method"] == "GET"
        assert data["path"] == "/"
        assert data["query_string"] == "param=value"
        assert data["headers"]["X-Test"] == "test-value"


@pytest.mark.asyncio
async def test_quart_provider_request_post():
    async with provider_http_app(handle_request_post) as app:
        client = app.test_client()

        test_data = {"key": "value"}
        response = await client.post("/", json=test_data)

        assert response.status_code == 200
        data = await response.get_json()

        assert data["method"] == "POST"
        assert data["content_type"] == "application/json"
        assert data["data"] == test_data


@pytest.mark.asyncio
async def test_quart_provider_websocket():
    async with (
        provider_ws_app(handle_websocket) as app,
        app.test_client().websocket("/") as test_client,
    ):
        test_data = {"message": "hello"}
        await test_client.send_json(test_data)

        response = await test_client.receive_json()

        assert response["path"] == "/"
        assert "headers" in response
        assert response["echo"] == test_data



from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager
from unittest.mock import Mock

import pytest
from dishka import FromDishka, make_async_container
from quart import Quart, Websocket

from quart_dishka.container import ContainerMiddleware
from quart_dishka.extension import _inject_routes, inject
from .mocks import (
    APP_DEP_VALUE,
    REQUEST_DEP_VALUE,
    WS_DEP_VALUE,
    AppDep,
    AppProvider,
    RequestDep,
    WebSocketAppProvider,
    WebSocketDep,
)


@asynccontextmanager
async def dishka_http_app(handler, provider) -> AsyncGenerator[Quart, None]:
    app = Quart(__name__)
    app.get("/")(inject(handler))
    container = make_async_container(provider)

    middleware = ContainerMiddleware(container)
    middleware.setup(app)

    try:
        yield app
    finally:
        await container.close()


@asynccontextmanager
async def dishka_ws_app(handler, provider) -> AsyncGenerator[Quart, None]:
    app = Quart(__name__)
    app.websocket("/")(inject(handler))
    container = make_async_container(provider)

    middleware = ContainerMiddleware(container)
    middleware.setup(app)

    try:
        yield app
    finally:
        await container.close()


@asynccontextmanager
async def dishka_auto_app(view, provider):
    app = Quart(__name__)
    container = make_async_container(provider)

    middleware = ContainerMiddleware(container)
    middleware.setup(app)

    _inject_routes(app)
    for blueprint in app.blueprints.values():
        _inject_routes(blueprint)

    app.route("/")(inject(view))
    yield app
    await container.close()


async def handle_with_app(
        a: FromDishka[AppDep],
        mock: FromDishka[Mock],
) -> None:
    mock(a)


@pytest.mark.parametrize("app_factory", [
    dishka_http_app, dishka_auto_app,
])
@pytest.mark.asyncio
async def test_http_app_dependency(app_provider: AppProvider, app_factory):
    async with app_factory(handle_with_app, app_provider) as app:
        test_client = app.test_client()
        await test_client.get("/")
        app_provider.mock.assert_called_with(APP_DEP_VALUE)
        app_provider.app_released.assert_not_called()
    app_provider.app_released.assert_called()


async def handle_with_request(
        a: FromDishka[RequestDep],
        mock: FromDishka[Mock],
) -> None:
    mock(a)


@pytest.mark.asyncio
async def test_http_request_dependency(app_provider: AppProvider):
    async with dishka_http_app(handle_with_request, app_provider) as app:
        test_client = app.test_client()
        await test_client.get("/")
        app_provider.mock.assert_called_with(REQUEST_DEP_VALUE)
        app_provider.request_released.assert_called_once()


@pytest.mark.asyncio
async def test_http_request_dependency2(app_provider: AppProvider):
    async with dishka_http_app(handle_with_request, app_provider) as app:
        test_client = app.test_client()
        await test_client.get("/")
        app_provider.mock.assert_called_with(REQUEST_DEP_VALUE)
        app_provider.request_released.assert_called_once()
        app_provider.mock.reset_mock()
        app_provider.request_released.reset_mock()
        await test_client.get("/")
        app_provider.mock.assert_called_with(REQUEST_DEP_VALUE)
        app_provider.request_released.assert_called_once()


@pytest.fixture(autouse=True)
def reset_mock(ws_app_provider) -> None:
    ws_app_provider.mock.reset_mock()
    ws_app_provider.app_released.reset_mock()
    ws_app_provider.request_released.reset_mock()
    ws_app_provider.websocket_released.reset_mock()


async def get_with_app(
    ws: FromDishka[Websocket],
    app_dep: FromDishka[AppDep],
    mock: FromDishka[Mock],
) -> None:
    await ws.accept()
    await ws.receive()  # consume the message
    mock(app_dep)
    await ws.send("passed")


@pytest.mark.asyncio
async def test_websocket_app_dependency(
    ws_app_provider: WebSocketAppProvider,
) -> None:
    async with (
        dishka_ws_app(get_with_app, ws_app_provider) as app,
        app.test_client().websocket("/") as test_client,
    ):
        await test_client.send("ping")
        assert await test_client.receive() == "passed"
        ws_app_provider.mock.assert_called_with(APP_DEP_VALUE)
        ws_app_provider.app_released.assert_not_called()


async def get_with_request(
    ws: FromDishka[Websocket],
    req_dep: FromDishka[RequestDep],
    mock: FromDishka[Mock],
) -> None:
    await ws.accept()
    await ws.receive()
    mock(req_dep)
    await ws.send("passed")


@pytest.mark.asyncio
async def test_websocket_request_dependency(
    ws_app_provider: WebSocketAppProvider,
) -> None:
    async with dishka_ws_app(get_with_request, ws_app_provider) as app:
        async with app.test_client().websocket("/") as test_client:
            await test_client.send("ping")
            assert await test_client.receive() == "passed"
            ws_app_provider.mock.assert_called_with(REQUEST_DEP_VALUE)
        ws_app_provider.request_released.assert_called_once()


@pytest.mark.asyncio
async def test_websocket_request_dependency_multiple(
    ws_app_provider: WebSocketAppProvider,
) -> None:
    async with dishka_ws_app(get_with_request, ws_app_provider) as app:
        async with app.test_client().websocket("/") as test_client:
            await test_client.send("ping")
            assert await test_client.receive() == "passed"
        ws_app_provider.request_released.assert_called_once()
        ws_app_provider.request_released.reset_mock()

        async with app.test_client().websocket("/") as test_client:
            await test_client.send("ping")
            assert await test_client.receive() == "passed"
        ws_app_provider.mock.assert_called_with(REQUEST_DEP_VALUE)
        ws_app_provider.request_released.assert_called_once()


async def get_with_websocket(
    ws: FromDishka[Websocket],
    ws_dep: FromDishka[WebSocketDep],
    mock: FromDishka[Mock],
) -> None:
    await ws.accept()
    await ws.receive()
    mock(ws_dep)
    await ws.send("passed")


@pytest.mark.asyncio
async def test_websocket_dependency(
    ws_app_provider: WebSocketAppProvider,
) -> None:
    async with dishka_ws_app(get_with_websocket, ws_app_provider) as app:
        async with app.test_client().websocket("/") as test_client:
            await test_client.send("ping")
            assert await test_client.receive() == "passed"
            ws_app_provider.mock.assert_called_with(WS_DEP_VALUE)
        ws_app_provider.websocket_released.assert_called_once()



from dataclasses import dataclass


class QuartDishkaError(Exception):
    """Base class for all QuartDishka exceptions"""


@dataclass(eq=False)
class ContainerNotSetError(QuartDishkaError):
    """Exception raised when the container is not set"""

    message: str = "Container must be set before initializing app"

    def __init__(self) -> None:
        super().__init__(self.message)

    @property
    def title(self) -> str:
        return self.message



import pytest
from dishka import FromDishka, make_async_container
from quart import Blueprint, Quart

from quart_dishka.exceptions import ContainerNotSetError
from quart_dishka.extension import QuartDishka, inject
from .mocks import APP_DEP_VALUE, AppDep, AppProvider


@pytest.mark.asyncio
async def test_quart_dishka_init():
    app = Quart(__name__)
    container = make_async_container(AppProvider())
    extension = QuartDishka(app=app, container=container)

    assert extension.container == container
    assert extension.auto_inject is False
    assert "QUART_DISHKA" in app.extensions


@pytest.mark.asyncio
async def test_quart_dishka_init_without_container():
    app = Quart(__name__)
    extension = QuartDishka(app=None)

    with pytest.raises(ContainerNotSetError,
                       match="Container must be set before initializing app"):
        extension.init_app(app)


@pytest.mark.asyncio
async def test_factory_pattern():
    container = make_async_container(AppProvider())
    extension = QuartDishka(container=container)

    app = Quart(__name__)
    extension.init_app(app)

    assert "QUART_DISHKA" in app.extensions


@pytest.mark.asyncio
async def test_quart_dishka_auto_inject():
    app = Quart(__name__)
    container = make_async_container(AppProvider())

    @app.route("/")
    async def index(app_dep: FromDishka[AppDep]) -> str:
        return str(app_dep)

    @app.route("/explicit")
    @inject
    async def explicit(app_dep: FromDishka[AppDep]) -> str:
        return str(app_dep)

    QuartDishka(app=app, container=container, auto_inject=True)

    client = app.test_client()

    response = await client.get("/")
    assert response.status_code == 200
    assert await response.get_data(as_text=True) == str(APP_DEP_VALUE)

    response = await client.get("/explicit")
    assert response.status_code == 200
    assert await response.get_data(as_text=True) == str(APP_DEP_VALUE)


@pytest.mark.asyncio
async def test_quart_dishka_blueprint_auto_inject():
    app = Quart(__name__)
    bp = Blueprint("test", __name__)
    container = make_async_container(AppProvider())

    @bp.route("/bp")
    async def bp_route(app_dep: FromDishka[AppDep]) -> str:
        return str(app_dep)

    @bp.route("/bp-explicit")
    @inject
    async def bp_explicit(app_dep: FromDishka[AppDep]) -> str:
        return str(app_dep)

    app.register_blueprint(bp)
    QuartDishka(app=app, container=container, auto_inject=True)

    client = app.test_client()

    response = await client.get("/bp")
    assert response.status_code == 200
    assert await response.get_data(as_text=True) == str(APP_DEP_VALUE)

    response = await client.get("/bp-explicit")
    assert response.status_code == 200
    assert await response.get_data(as_text=True) == str(APP_DEP_VALUE)
